/* Mnplane.java by Wolf Jung (C) 2010.  www.mndynamics.com
   Defines the Thread class Mnplane.

   This class is part of Juliette 1.0, which is free software; you can
   redistribute and / or modify them under the terms of the GNU General Public
   License as published by the Free Software Foundation, see www.gnu.org;
   either version 3, or (at your option) any later version. In short: there is
   no warranty of any kind; you must redistribute the source code as well.
*/
import java.awt.*;
import java.awt.image.*;

/* An Mnplane computes an off-screen BufferedImage, it has no direct visual
   representation. The current point is X, Y and hmid, vmid, pw, ph describe
   the plane. The parameter is A, B and A > 1.0e10 means not dynamic plane.
   as* and gamma, ra, ia, rlogrho, ilogrho, rk, ik refer to scaling at
   a Misiurewicz point. Preper, Per, temp are used in renormalization.
   The thread cannot be restarted, so it is stopped and awakened by
   draw() using the volatile execute to synchronize. Since buffer will
   be used to draw the parent, access to it is synchronized as well.

   * Drawing a rescaled version of the old image upon zooming did
     not work yet, probably because of an asynchronous operation.
*/
public class Mnplane extends Thread
{  //3 stop, 2 work, 1 idle, 0 working, -1 kill, -2 killed
   protected volatile int execute;
   protected int imax, kmax, asn, aspre, asper;
   protected short nmax, Preper, Per;
   protected double X, Y, A, B, hmid, vmid, pw, ph,
      gamma, ra, ia, rlogrho, ilogrho, rk, ik, temp;
   protected BufferedImage buffer;
   protected Graphics2D g;
   protected Component parent;

Mnplane(int Imax, int Kmax, BufferedImage aBuffer, Component Parent)
{  execute = 1; imax = Imax; kmax = Kmax;
   buffer = aBuffer; parent = Parent;
   nmax = 100; Preper = 0; Per = 0;
   asn = 0; aspre = 0; asper = 0; gamma = 0; ra = 0; ia = 0;
   rlogrho = 0; ilogrho = 0; rk = 0; ik =0;
   X = 0; Y = 0; A = 1.0e100; B = 0; temp = 0;
   hmid = 0; vmid = 0; pw = 1.0; ph = 0;
}

public void setPos(int[] ik)
{  ik[0] -= imax; ik[1] -= kmax;
   X = hmid + ik[0]*pw + ik[1]*ph; Y = vmid + ik[0]*ph - ik[1]*pw;
}

public void getPos(int[] ik)
{ getPos(ik, X, Y); }

public void getPos(int[] ik, double x, double y)
{  double pdet = pw*pw + ph*ph,
   I = Math.floor(((x - hmid)*pw + (y - vmid)*ph)/pdet),
   K = Math.floor(((x - hmid)*ph + (vmid - y)*pw)/pdet);
   if (I < -imax-4 || I >= imax+4 || K < -kmax-4 || K >= kmax+4)
   { ik[0] = -10; ik[1] = -10; }
   else
   { ik[0] = imax + (int) I; ik[1] = kmax  + (int) K; }
}

public void move(int i, int k)
{  int[] ik = new int[2]; getPos(ik);
   if (ik[0] <= -10 || (i == 0 && k == 0) )
   { X = hmid; Y = vmid; }
   else
   { X += (i*pw - k*ph); Y += (i*ph + k*pw); }
}

public void setPoint(double[] z)
{ X = z[0]; Y = z[1]; }

public void getPoint(double[] z)
{ z[0] = X; z[1] = Y; }

public void setParameter(double[] c)
{ Stop(); A = c[0]; B = c[1]; }

public void getParameter(double[] c)
{ c[0] = A; c[1] = B; }

public void setNmax(int n)
{ Stop(); nmax = (short) n; }

public int getNmax()
{ return (int) nmax; }

public void setPlane(double[] t)
{ Stop(); hmid = t[0]; vmid = t[1]; pw = t[2]/imax; ph = t[3]/imax; }

public void getPlane(double[] t)
{ t[0] = hmid; t[1] = vmid; t[2] = pw*imax; t[3] = ph*imax; }

public void iterate()
{  if (A > 1.0e10) return;
   double u = X*X, v = Y*Y; if (u + v > 1.0e100) return;
   Y = 2*X*Y + B; X = u - v + A;
}

public synchronized void Stop()
{  if (execute != 0) return;
   execute = 3;
   while (execute > 1)
      try { wait(); } catch (InterruptedException e) {}
}

public synchronized void Destroy()
{  if (execute < 0) return;
   execute = -1;
   while (execute > -2)
      try { wait(); } catch (InterruptedException e) {}
}

public synchronized void newtonRay(long N, long D)
{  if (execute == 0)
   {  execute = 3;
      while (execute > 1)
         try { wait(); } catch (InterruptedException e) {}
   }
   if (execute != 1) return;
   g.setColor(Color.white);
   int q = 5, j; int[] ik0 = new int[2]; int[] ik = new int[2];
   double logR = 14.0, u; double[] z = new double[2];
   u = Math.exp(0.5*logR); z[1] = N*2.0*Math.PI/D;
   z[0] = u*Math.cos(z[1]); z[1] = u*Math.sin(z[1]); getPos(ik0, z[0], z[1]);
synchronized (buffer) {
   for (Per = 1; Per <= nmax + 2; Per++)
   {  N *= 2; if (N >= D) N -= D;
      for (j = 1; j <= q; j++)
      {  if (rayNewton(z, Math.exp(-j*0.69315/q)*logR, N*2.0*Math.PI/D) != 0)
         { Per = 32500; break; }
         getPos(ik, z[0], z[1]);
         if (ik0[0] >= 0)
         {  if (ik[0] >= 0) g.drawLine(ik0[0], ik0[1], ik[0], ik[1]);
            else { Per = 32500; break; }
         }
         ik0[0] = ik[0]; ik0[1] = ik[1];
      }
   }
}  parent.repaint();
} //newtonRay

public int rayNewton(double[] z, double rlog, double ilog)
{  int k, l; double x = z[0], y = z[1], fx, fy, px, py, u, v = 0,
      a = A, b = B, d = 1.0 + x*x + y*y, t0, t1;
   for (k = 1; k <= 60; k++)
   {  if (A > 1.0e10) { a = x; b = y; }
      fx = Math.cos(ilog); fy = Math.sin(ilog);
      t0 = Math.exp(rlog)*fx - 0.5*Math.exp(-rlog)*(a*fx + b*fy);
      t1 = Math.exp(rlog)*fy - 0.5*Math.exp(-rlog)*(b*fx - a*fy);
      fx = x; fy = y; px = 1.0; py = 0;
      for (l = 1; l <= Per; l++)
      {  u = 2.0*(fx*px - fy*py); py = 2.0*(fx*py + fy*px);
         px = u; if (A > 1.0e10) px++;
         u = fx*fx; v = fy*fy; fy = 2.0*fx*fy + b; fx = u - v + a;
         u += v; v = px*px + py*py; if (u + v > 1.0e100) return 1;
      }
      fx -= t0; fy -= t1; if (v < 1.0e-50) return 2;
      u = (fx*px + fy*py)/v; v = (fx*py - fy*px)/v;
      px = u*u + v*v; if (px > 9.0*d) { z[0] = x; z[1] = y; return -1; }
      x -= u; y += v; d = px; if (px < 1.0e-26) break;
   }
   z[0] = x; z[1] = y; return 0;
} //rayNewton

public synchronized void draw(double factor, int preper, int per)
{  if (execute < 0 || execute > 1 || buffer == null || factor <= 0) return;
   if (factor > 1.0 && pw*pw + ph*ph > 0.01) return;
   if (g == null) g = buffer.createGraphics();
   if (execute == 0)
   {  execute = 3;
      while (execute > 1)
         try { wait(); } catch (InterruptedException e) {}
   }
   Preper = (short) preper; Per = (short) per;
   if (Per > 0) { temp = factor; factor = 1.0; if (temp > 0.3) temp = 0.25; }
   if (factor != 1.0)
   {  int[] ik = new int[2]; getPos(ik); if (ik[0] < 0) factor = 1.0;
      pw *= factor; ph *= factor;
      hmid = X + factor*(hmid - X); vmid = Y + factor*(vmid - Y);
   }
   execute = 2; notifyAll();
}

public void run ()
{  int i, k, n, u, v; double x, y;
   short[] tempo; short[] last = new short[2*imax+2];
   short[] now = new short[2*imax+2]; short[] next = new short[2*imax+2];
while(true){
   synchronized (this)
   {  while (execute  == 1)
         try { wait(); } catch (InterruptedException e) {}
      if (execute < 0) { execute = -2; notifyAll(); return; }
      execute = 0;
   }
   if (Per > 0)
   {  if (Preper <= 0) Preper = Per;
      g.setColor(Color.magenta);
   }
   x = hmid - (kmax+1)*ph; y = vmid + (kmax+1)*pw;
   for (i = -imax-1; i <= imax; i++)
     last[i+imax+1] = esctime(x + i*pw, y + i*ph);
   x = hmid - kmax*ph; y = vmid + kmax*pw;
   for (i = -imax-1; i <= imax; i++)
     now[i+imax+1] = esctime(x + i*pw, y + i*ph);
   for (k = -kmax; k < kmax; k++)
   {  if (execute != 0) break;
      x = hmid + (k+1)*ph; y = vmid - (k+1)*pw;
      for (i = -imax-1; i <= imax; i++)
        next[i+imax+1] = esctime(x + i*pw, y + i*ph);
      for (i = -imax; i < imax; i++)
      {  if (now[i+imax+1] > 32000) last[i+imax] = 0;
         else
         {  n = 0; u = now[i+imax+1];
            for (v = i+imax; v <= i+imax+2; v++)
            { if (last[v] < u) n++; if (next[v] < u) n++; }
            if (now[i+imax] < u) n++; if (now[i+imax+2] < u) n++;
            if (n > 4) last[i+imax] = 0;
          }
      } //for i
      synchronized (buffer)
      {  if (Per <= 0)
         {  g.setColor(Color.green);
            g.drawLine(0, k+kmax, 2*imax-1, k+kmax);
            g.setColor(Color.black);
         }
         for (i = 0; i < 2*imax; i++)
           if (last[i] == 0) g.drawLine(i, k+kmax, i, k+kmax);
      }
      tempo = last; last = now; now = next; next = tempo;
      if ((kmax - k) % 16 == 1) parent.repaint();
   } //for k
   synchronized (this)
   { if (execute >= 0) { execute = 1; notifyAll(); } }
}} //run

protected short esctime(double x, double y)
{  short j; double a, b, u, v, x0 = x, y0 = y;
   if (A < 1.0e10) { a = A; b = B; }
   else { a = x; b = y; }
   for (j = 1; j <= nmax; j++)
   {  u = x*x; v = y*y;
      if (u + v > 16.0) return j;
      y = 2*x*y + b; x = u - v + a;
      if (Per == 0) //check periods 1...4
      {  if ( (x - x0)*(x - x0) + (y - y0)*(y - y0) < 1e-7
              && pw*pw + ph*ph > 1e-8) break;
         if ((j & 3) == 0) { x0 = x; y0 = y; }
      }
      else //j - 1 = Preper - 1 + n*Per, renormalize
      { if (j >= Preper && (j-Preper) % Per == 0 && u + v > temp) return j; }
   }
   return 32001;
}

public int period(double[] c)
{  if (c == null || (c[0] == -2.0 && c[1] == 0)
       || (c[0] == 0 && c[1]*c[1] == 1.0) ) return 10000;
   int j; double x = 0, y = 0, x0, y0, u, v;
   for (j = 1; j <= 1000; j++)
   {  u = x*x; v = y*y; if (u + v > 4.0) return 0;
      y = 2.0*x*y + c[1]; x = u - v + c[0];
   }
   x0 = x; y0 = y;
   for (j = 1; j <= 1024; j++)
   {  u = x*x; v = y*y; if (u + v > 4.0) return 0;
      y = 2.0*x*y + c[1]; x = u - v + c[0];
      if ( (x - x0)*(x - x0) + (y - y0)*(y - y0) < 1e-24) return j;
   }
   return 10000;
} //period

public int find(int preper, int per, double[] z)
{  if (preper < 0 || per == 0 || z == null) return 5;
   int i, j; double x = z[0], y = z[1], a = A, b = B, fx, fy, px, py, w;
   if (per < 0) //multiplier for Misiurewicz
   {  per = -per; px = 1.0; py = 0;
      for (j = 1; j < preper + per; j++)
      {  fx = x*x; fy = y*y; if (fx + fy > 1.0e100) return 1;
         y = 2.0*x*y + z[1]; x = fx - fy + z[0];
         if (j >= preper)
         { w = 2.0*(x*px - y*py); py = 2.0*(x*py + y*px); px = w; }
      }
      z[0] = px; z[1] = py; return 0;
   }
   for (i = 0; i < 30; i++)
   {  if (A > 1.0e10) { a = x; b = y; }
      if (preper == 0)
      {  if (A > 1.0e10) { fx = 0; fy = 0; px = 0; py = 0; }
         else { fx = -x; fy = -y; px = -1; py = 0; }
      }
      else
      {  fx = x; fy = y; px = 1.0; py = 0;
         for (j = 1; j < preper; j++)
         {  if (px*px + py*py > 1.0e100) return 1;
            w = 2.0*(fx*px - fy*py); py = 2.0*(fx*py + fy*px);
            px = w; if (A > 1.0e10) px++;
            w = fx*fx - fy*fy + a; fy = 2.0*fx*fy + b; fx = w;
         }
      }
      double Fx = fx, Fy = fy, Px = px, Py = py;
      for (j = 0; j < per; j++)
      {  if (px*px + py*py > 1.0e100) return 2;
         w = 2.0*(fx*px - fy*py); py = 2.0*(fx*py + fy*px);
         px = w; if (A > 1.0e10) px++;
         w = fx*fx - fy*fy + a; fy = 2.0*fx*fy + b; fx = w;
      }
      fx += Fx; fy += Fy; px += Px; py += Py;
      w = px*px + py*py; if (w < 1.0e-100) return 3;
      x -= (fx*px + fy*py)/w; y += (fx*py - fy*px)/w;
   }
   z[0] = x; z[1] = y; return 0;
} //find

/* Returns lambda and zp when per < 0, otherwise the dynamic plane
   corresponding to the current parameter plane according to local
   similarity. It works with similar formulas in the Misiurewicz case.
*/
public int similarity(int preper, int per, double[] t)
{  if (preper < 0 || per == 0 || A < 1.0e10) return 1;
   boolean lambda = false; if (per < 0) { per = -per; lambda = true; }
   int i; t[0] = X; t[1]= Y; if (find(preper, per, t) != 0) return 1;
   double a = X, b = Y, x = t[0], y = t[1], xp = 1.0, yp = 0, xd = 1.0, yd = 0,
      xp0 = 0, yp0 = 0, xd0 = 0, yd0 = 0, u;
   for (i = 1; i < preper + per; i++)
   {  if (i == preper) { xp0 = xp; yp0 = yp; xd0 = xd; yd0 = yd; }
      u = 2*(x*xp - y*yp); yp = 2*(x*yp + y*xp); xp = u;
      u = 2*(x*xd - y*yd); yd = 2*(x*yd + y*xd); xd = u + 1.0;
      u = x*x - y*y + t[0]; y = 2*x*y + t[1]; x = u;
      if (x*x + y*y + xp*xp + yp*yp + xd*xd + yd*yd > 1.0e100) return 2;
      if (preper == 0 && i <  per - 1 && x*x + y*y < 1.0e-24) return 3; //p<per
      if (preper == 0 && i == per - 1 && x*x + y*y > 1.0e-20) return 4; //p>per
   }
   xp += xp0; yp += yp0; xd += xd0; yd += yd0;
   yd0 = xp*xp + yp*yp; if (yd0 < 1.0e-100) return 5;
   u = (xp*xd + yp*yd)/yd0; yd = (xp*yd - yp*xd)/yd0; xd = u; //zd = lambda
   if (lambda) { t[0] = xd; t[1] = yd; t[2] = xp; t[3] = yp; return 0; }
   getPlane(t);
   if (preper > 0)
   {  t[0] -= a; t[1] -= b;
      u = a + xd*t[0] - yd*t[1]; t[1] = b + xd*t[1] + yd*t[0]; t[0] = u;
      u = xd*t[2] - yd*t[3]; t[3] = xd*t[3] + yd*t[2]; t[2] = u;
      return 0;
   }
   u = xp*xp - yp*yp; yp = 2*xp*yp; xp = u; //zp = A
   yd0 = 2*(xp*xp + yp*yp); if (yd0 < 1.0e-100) return 6;
   x = xp/yd0; y = -yp/yd0; //z = 1/2A
   t[0] -= a; t[1] -= b;
   u = a + xd*t[0] - yd*t[1] + x; t[1] = b + xd*t[1] + yd*t[0] + y; t[0] = u;
   u = xd*t[2] - yd*t[3]; t[3] = xd*t[3] + yd*t[2]; t[2] = u;
   return 0;
} //similarity

/* Sets the parameters for asymptotic scaling at a Misiurerwicz point.
   Here preiter will be greater than the preperiod, it is the smallest
   period of a little Mandelbrot set, or the number of iterations from
   a neighborhood of the critical value to a superset.
*/
public void setScale(int preiter, int per, double[] t)
{  if (per <= 0 || A < 1.0e10) return;
   ra = t[0]; ia = t[1]; rlogrho = Math.log(t[2]); ilogrho = t[3];
   gamma = 3.0; asn = 0; aspre = preiter; asper = per;
   if (preiter > 0) { rk = t[4]; ik = t[5]; }
   else { rk = imax*pw/1.3; ik = imax*ph/1.3; }
}

/* Returns the period of c_n when Gamma < 0, sets the starting plane
   when Gamma > 0, and increases n++ when Gamma = 0.
*/
public int scale(double Gamma)
{  if (asper <= 0 || A < 1.0e10) return 0;
   if (Gamma < 0) //period
   { if (aspre <= 0 || gamma > 2.0) return 0; else return aspre + asper*asn; }
   Stop(); double[] t = new double[4];
   if (Gamma > 0) //start
   {  asn = 0; gamma = Gamma; if (aspre <= 0) gamma = 3.0;
      if (gamma > 2.0)
      { t[0] = ra; t[1] = ia; t[2] = 1.3*Math.sqrt(rk*rk + ik*ik); }
      else
      {  t[0] = ra + rk; t[1] = ia + ik; find(0, aspre, t);
         t[2] = (2.3 - gamma)*Math.sqrt(rk*rk + ik*ik);
      }
      t[3] = 0; setPlane(t); setPoint(t); return 0;
   }
   asn++; getPlane(t); double[] z = {X, Y};
   if (gamma > 2.0) //around a
   {  z[0] -= ra; z[1] -= ia; expmult(-1.0, z); X = z[0] + ra; Y = z[1] + ia;
      t[0] -= ra; t[1] -= ia; expmult(-1.0, t); t[0] += ra; t[1] += ia;
      z[0] = t[2]; z[1] = t[3]; expmult(-1.0, z); t[2] = z[0]; t[3] = z[1];
   }
   else //else around c_n
   {  double[] c = {rk, ik}; expmult(1.0-asn, c); c[0] += ra; c[1] += ia;
      find(0, aspre + (asn-1)*asper, c); //old c_n
      double[] C = {rk, ik}; expmult(-asn, C); C[0] += ra; C[1] += ia;
      find(0, aspre + asn*asper, C); //new c_n
      z[0] -= c[0]; z[1]-= c[1]; expmult(-gamma, z);
      X = z[0] + C[0]; Y = z[1] + C[1];
      t[0] -= c[0]; t[1]-= c[1]; expmult(-gamma, t); t[0] += C[0]; t[1] += C[1];
      z[0] = t[2]; z[1] = t[3]; expmult(-gamma, z); t[2] = z[0]; t[3] = z[1];
   }
   setPlane(t); return 0;
} //scale

protected void expmult(double t, double[] z)
{  //multiply  x + iy  with  rho^t
   double u = Math.cos(t*ilogrho), v = Math.sin(t*ilogrho), w;
   w = u*z[0] - v*z[1]; z[1] = u*z[1] + v*z[0];
   u = Math.exp(t*rlogrho); z[0] = u*w; z[1] *= u;
}

} //Mnplane
